import bz2
import pickle
import os
import pandas as pd

from tqdm import tqdm
from wikidatasets.utils import get_results, clean
from wikidatasets.utils import get_pickle_path, write_to_pickle
from wikidatasets.utils import get_id, get_label, to_triplets, intersect, to_json
from wikidatasets.utils import concatpkls, write_csv, write_ent_dict, write_rel_dict, write_readme, relabel


def get_subclasses(subject):
    """Get a list of WikiData IDs of entities which are subclasses of the subject.

    Parameters
    ----------
    subject: str
        String describing the subject (e.g. 'Q5' for human).

    Returns
    -------
    result: list
        List of WikiData IDs of entities which are subclasses of the subject.

    """
    endpoint_url = "https://query.wikidata.org/sparql"
    query = """SELECT ?item WHERE {?item wdt:P279* wd:""" + subject + """ .}"""

    results = get_results(endpoint_url, query)

    return [clean(result['item']['value']) for result in results['results']['bindings']]


def query_wikidata_dump(dump_path, path, n_lines, test_entities=None, collect_labels=False):
    """This function goes through a Wikidata dump. It can either collect entities that are instances of \
    `test_entities` or collect the dictionary of labels. It can also do both.

    Parameters
    ----------
    dump_path: str
        Path to the latest-all.json.bz2 file downloaded from https://dumps.wikimedia.org/wikidatawiki/entities/.
    path: str
        Path to where pickle files will be written.
    n_lines: int
        Number of lines of the dump. Fastest way I found was `$ bzgrep -c ".*" latest-all.json.bz2`.
        This can be an upper-bound as it is only used for displaying a progress bar.
    test_entities: list
        List of entities to check if a line is instance of. For each line (entity), we check if it as a fact of the \
        type (id, query_rel, test_entity).
    collect_labels: bool
        Boolean indicating whether the labels dictionary should be collected.

    """
    pickle_path = get_pickle_path(path)
    collect_facts = (test_entities is not None)
    fails = []

    if collect_labels:
        labels = {}
    if collect_facts:
        facts = []
        n_pickle_dump = 0

    dump = bz2.open(dump_path, 'rt')
    progress_bar = tqdm(total=n_lines)
    counter = 0  # counter of the number of lines read
    line = dump.readline()  # the first line of the file should be "[\n" so we skip it

    while True:
        # while there are lines to read
        line = dump.readline().strip()
        if len(line) == 0:
            break

        counter += 1
        progress_bar.update(1)

        try:
            line = to_json(line)

            if collect_labels:
                id_ = get_id(line)
                labels[id_] = get_label(line)

            if collect_facts:
                triplets, instanceOf = to_triplets(line)
                if len(instanceOf) > 0 and intersect(instanceOf, test_entities):
                    facts.extend(triplets)

        except:
            if type(line) == dict and ('claims' in line.keys()):
                if len(line['claims']) != 0:
                    fails.append(line)
            else:
                fails.append(line)

        if counter % 3000000 == 0:
            # dump in pickle to free memory
            if collect_facts:
                n_pickle_dump += 1
                facts, fails = write_to_pickle(pickle_path, facts, fails, n_pickle_dump)

    if collect_facts:
        _, _ = write_to_pickle(pickle_path, facts, fails, n_pickle_dump + 1)
    if collect_labels:
        pickle.dump(labels, open(path + 'labels.pkl', 'wb'))


def build_dataset(path, labels, return_=False, dump_date='23rd April 2019'):
    """Builds datasets from the pickle files produced by the query_wikidata_dump.

    Parameters
    ----------
    path: str
        Path to the directory where there should already be a pickles/ directory. In the latter directory, all \
        the .pkl files will be concatenated into one dataset.
    labels: dict
        Dictionary collected by the query_wikidata_dump function when collect_labels is set to True.
    return_: bool
        Boolean indicating if the built dataset should be returned on top of being written on disk.
    dump_date: str
        String indicating the date of the Wikidata dump used. It is used in the readme of the dataset.

    Returns
    -------
    edges: pandas.DataFrame
        DataFrame containing the edges between entities of the graph.
    attributes: pandas.DataFrame
        DataFrame containing edges linking entities to their attributes.
    entities: pandas.DataFrame
        DataFrame containing a list of all entities & attributes with their Wikidata IDs and labels.
    relations: pandas.DataFrame
        DataFrame containing a list of all relations with their Wikidata IDs and labels.
    """

    if path[-1] != '/':
        path = path+'/'
    path_pickle = path + 'pickles/'
    n_files = len([name for name in os.listdir(path_pickle) if name[-4:] == '.pkl'])
    df = concatpkls(n_files, path_pickle)

    ents = list(df['headEntity'].unique())
    feats = list(set(df['tailEntity'].unique()) - set(ents))
    ent2ix = {ent: i for i, ent in enumerate(ents + feats)}
    ix2ent = {i: ent for ent, i in ent2ix.items()}

    tmp = df['relation'].unique()
    rel2ix = {rel: i for i, rel in enumerate(tmp)}
    ix2rel = {i: rel for rel, i in rel2ix.items()}

    df['headEntity'] = df['headEntity'].apply(lambda x: ent2ix[x])
    df['tailEntity'] = df['tailEntity'].apply(lambda x: ent2ix[x])
    df['relation'] = df['relation'].apply(lambda x: rel2ix[x])

    nodes = pd.DataFrame([[i, ix2ent[i]] for i in range(len(ents))],
                         columns=['entityID', 'wikidataID'])
    nodes['label'] = nodes['wikidataID'].apply(relabel, args=(labels,))

    entities = pd.DataFrame([[i, ix2ent[i]] for i in range(len(ix2ent))],
                            columns=['entityID', 'wikidataID'])
    entities['label'] = entities['wikidataID'].apply(relabel, args=(labels,))

    relations = pd.DataFrame([[i, ix2rel[i]] for i in range(len(ix2rel))],
                             columns=['relationID', 'wikidataID'])
    relations['label'] = relations['wikidataID'].apply(relabel, args=(labels,))

    edges_mask = df.tailEntity.isin(df['headEntity'].unique())
    edges = df.loc[edges_mask, ['headEntity', 'tailEntity', 'relation']]
    attributes = df.loc[~edges_mask, ['headEntity', 'tailEntity', 'relation']]

    write_csv(edges, path + 'edges.tsv')
    write_csv(attributes, path + 'attributes.tsv')
    write_ent_dict(nodes, path + 'nodes.tsv')
    write_ent_dict(entities, path + 'entities.tsv')
    write_rel_dict(relations, path + 'relations.tsv')
    write_readme(path+'readme.md',
                 n_core_ents=attributes['headEntity'].nunique(),
                 n_attrib_ents=attributes['tailEntity'].nunique(),
                 n_core_rels=edges['relation'].nunique(),
                 n_attrib_rels=attributes['relation'].nunique(),
                 n_core_facts=len(edges),
                 n_attrib_facts=len(attributes),
                 dump_date=dump_date)

    if return_:
        return edges, attributes, entities, relations



# -*- coding: utf-8 -*-

"""Top-level package for WikiDataSets."""

__author__ = """Armand Boschin"""
__email__ = 'aboschin@enst.fr'
__version__ = '0.3.0'

from .processFunctions import get_subclasses
from .processFunctions import query_wikidata_dump
from .processFunctions import build_dataset

from .utils import load_data_labels



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ['tqdm>=4.43', 'sparqlwrapper>=1.8.5', 'pandas>=1.0.3']

setup_requirements = []

test_requirements = []

setup(
    author="Armand Boschin",
    author_email='aboschin@enst.fr',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    description="Break WikiData dumps into smaller knowledge graphs",
    install_requires=requirements,
    license="BSD license",
    long_description=readme,
    include_package_data=True,
    keywords='wikidatasets',
    name='wikidatasets',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/armand33/wikidatasets',
    version='0.3.0',
    zip_safe=False,
)



import pickle
import json
import pandas as pd
import os

from wikidatasets.exceptions import ParsingException
from tqdm import tqdm

from SPARQLWrapper import SPARQLWrapper, JSON


def get_results(endpoint_url, query):
    sparql = SPARQLWrapper(endpoint_url)
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    return sparql.query().convert()


def to_json(line):
    if line[-1] == ',':
        line = line[:-1]  # all lines should end with a ','

    # turn string to json
    if line[0] != '{' or line[-1] != '}':
        # then this line is not a proper json file we should deal with it later
        raise ParsingException

    return json.loads(line)


def concat_claims(claims):
    """

    Parameters
    ----------
    claims: dict

    Returns
    -------
    iter: iterator
        iterator through the claims
    """

    for rel_id, rel_claims in claims.items():
        for claim in rel_claims:
            yield claim


def to_triplets(ent):
    """

    Parameters
    ----------
    ent: dict
        Dictionary coming from the parsing of a json line of the dump.

    Returns
    -------
    triplets: list
        List of triplets of this entity (head, rel, tail).
    """
    if len(ent['claims']) == 0:
        return []
    claims = concat_claims(ent['claims'])
    triplets = []
    instanceof = []
    e1 = ent['id']
    for claim in claims:
        mainsnak = claim['mainsnak']
        if mainsnak['snaktype'] != "value":
            continue
        if mainsnak['datatype'] == 'wikibase-item':
            rel = mainsnak['property']
            e2 = 'Q{}'.format(mainsnak['datavalue']['value']['numeric-id'])
            triplets.append((e1, rel, e2))
            if rel == 'P31':
                instanceof.append(e2)
    return triplets, instanceof


def get_type(ent):
    return ent['type']


def get_id(ent):
    return ent['id']


def get_label(ent):
    """

    Parameters
    ----------
    ent: dict
        Dictionary coming from the parsing of a json line of the dump.

    Returns
    -------
    label: str
        Label of ent in english if available of any other language else.
    """

    labels = ent['labels']
    if len(labels) == 0:
        return 'No label {}'.format(ent['id'])
    if 'en' in labels.keys():
        return labels['en']['value']
    else:
        return labels[list(labels.keys())[0]]['value']


def relabel(x, labels):
    try:
        lab = labels[x]
        if ':' in lab:
            return lab[lab.index(':')+1:]
        else:
            return lab
    except KeyError:
        return x


def clean(str_):
    if str_[:31] == 'http://www.wikidata.org/entity/':
        return str_[31:]
    else:
        print('problem')
        return ''


def get_pickle_path(path):
    if path[-1] != '/':
        path = path+'/'
    pickle_path = path + 'pickles/'
    if not os.path.exists(pickle_path):
        os.makedirs(pickle_path)
    return pickle_path


def write_to_pickle(pickle_path, facts, fails, n_pickle_dump):
    pickle.dump((facts, fails),
                open(pickle_path + 'dump{}.pkl'.format(n_pickle_dump), 'wb'))
    print('Just made pickle dump number {}'.format(n_pickle_dump))
    return [], []


def intersect(long_list, short_list):
    return len(set(long_list).intersection(set(short_list))) > 0


def count_true_fails(fails):
    true_fails = 0
    for f in fails:
        try:
            if str(f) == ']':
                continue  # in this case it's the last line of the original dump file
            if len(f['claims']) > 0:
                true_fails += 1
        except:
            print(f)
    return true_fails


def concatpkls(n_dump, path_pickle, labels=None):
    df = pd.DataFrame(columns=['headEntity', 'relation', 'tailEntity'])

    for nd in tqdm(range(n_dump)):
        with open(path_pickle + 'dump{}.pkl'.format(nd + 1), 'rb') as f:
            facts, fails = pickle.load(f)
            true_fails = count_true_fails(fails)
            if true_fails > 0:
                print('{} true fails'.format(true_fails))
        df = pd.concat([df, pd.DataFrame(facts, columns=['headEntity', 'relation', 'tailEntity'])])
    df = df.drop_duplicates()

    if labels is not None:
        df['headEntity'] = df['headEntity'].apply(relabel, args=(labels,))
        df['relation'] = df['relation'].apply(relabel, args=(labels,))
        df['tailEntity'] = df['tailEntity'].apply(relabel, args=(labels,))

    return df


def write_csv(df, name):
    with open(name, 'w', encoding='utf-8') as f:
        f.write('headEntity\ttailEntity\trelation\n')
        df.to_csv(f, sep='\t', header=False, index=False)


def write_ent_dict(df, name):
    with open(name, 'w', encoding='utf-8') as f:
        f.write('entityID\twikidataID\tlabel\n')
        df.to_csv(f, sep='\t', header=False, index=False)


def write_rel_dict(df, name):
    with open(name, 'w', encoding='utf-8') as f:
        f.write('relationID\twikidataID\tlabel\n')
        df.to_csv(f, sep='\t', header=False, index=False)


def write_readme(name, n_core_ents, n_attrib_ents,
                 n_core_rels, n_attrib_rels,
                 n_core_facts, n_attrib_facts,
                 dump_date):
    with open(name, 'w', encoding='utf-8') as f:
        f.write('Built with WikiData dump from {}.\n'.format(dump_date))
        f.write("Here are some meta data about this data set:\n")
        f.write("Core entities: {}\n".format(n_core_ents))
        f.write("Attribute entities: {}\n".format(n_attrib_ents))
        f.write("Core relations: {} (number of different relations involving only core entities)\n".format(n_core_rels))
        f.write("Attribute relations: {} (number of different relations from core entities "
                "to attribute entities)\n".format(n_attrib_rels))
        f.write("Core facts: {} (facts involving only core entities)\n".format(n_core_facts))
        f.write("Attribute facts : {} (facts linking core entities "
                "to their attribute entities)\n".format(n_attrib_facts))
        f.write("Find more details about this dataset at https://arxiv.org/abs/1906.04536.")


def load_data_labels(path, attributes=False, return_dicts=False):
    """This function loads the edges and attributes files into Pandas dataframes and merges the labels of entities and \
    relations to get.

    Parameters
    ----------
    path: str
        Path to the directory containing the edges.txt, attributes.txt, entities.txt, relations.txt files.
    attributes: bool
        Boolean indicating if we should read the attributes files. If False, then the edges file is read.
    return_dicts: bool
        Boolean indicating if the entities and relations labels dictionaries should be returned.

    Returns
    -------
    df: pandas.DataFrame
        DataFrame containing either the edges or the attributes depending on the value of `attributes`.
    entities: pandas.DataFrame
        DataFrame containing the list of all entities and wikidata IDs and labels.
    relations: pandas.DataFrame
        DataFrame containing the list of all relations and wikidata IDs and labels.
    """

    if attributes:
        df = pd.read_csv(path + 'attributes.tsv', sep='\t')
    else:
        df = pd.read_csv(path + 'edges.tsv', sep='\t')

    entities = pd.read_csv(path + 'entities.tsv', sep='\t')
    relations = pd.read_csv(path + 'relations.tsv', sep='\t')

    df = pd.merge(left=df, right=entities[['entityID', 'label']], left_on='headEntity',
                  right_on='entityID')
    df.drop('entityID', axis=1, inplace=True)
    df.columns = list(df.columns[:-1]) + ['headLabel']

    df = pd.merge(left=df, right=entities[['entityID', 'label']], left_on='tailEntity',
                  right_on='entityID')
    df.drop('entityID', axis=1, inplace=True)
    df.columns = list(df.columns[:-1]) + ['tailLabel']

    df = pd.merge(left=df, right=relations[['relationID', 'label']], left_on='relation',
                  right_on='relationID')
    df.drop('relationID', axis=1, inplace=True)
    df.columns = list(df.columns[:-1]) + ['relationLabel']

    if return_dicts:
        return df, entities, relations

    return df
